/*
 * File:         cube2_90_down.c
 * Date:         Decemeber 7th, 2006
 * Description:  A controller moving four Cube2 straight and climbing down.
 * Author:       Simon Blanchoud
 * Modifications:
 */

#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <device/robot.h>
#include <device/custom_robot.h>
#include <device/servo.h>
#include <device/connector.h>
#include <device/emitter.h>
#include <device/receiver.h>

#define TIME_STEP 64
#define SERVO_NUMBER 3
#define CONNECTOR_NUMBER 24
#define FACES_NUMBER 6
#define TOL 0.000001

/* An example of possible states*/
typedef enum
{ ROTATING, DETACHING, ROTATING_LOCK, LISTENING, IDLE} States;
typedef enum { WITH, WITHOUT, ONLY } Behaviors;

static void reset(void);
static int run(int);

/* The functions used to interact with the roombot. 
 *
 * The second parameter for the functions presence, attach and detach are 
 * related to the connector of the corresponding face which is placed on
 * the rotating servo. Its value can be:
 *      - WITH : Take also the rotating connector into account
 *      - WITHOUT : Do not take it into account
 *      - ONLY : Take only this connector into account
 */
static int presence(int, Behaviors);
static void attach(int, Behaviors);
static void detach(int, Behaviors);
static void rotate(int, float);
static int in_place(int);

static DeviceTag servos[SERVO_NUMBER];
static DeviceTag connectors[CONNECTOR_NUMBER];
static DeviceTag emitters[FACES_NUMBER];
static DeviceTag receivers[FACES_NUMBER];

static double target_positions[SERVO_NUMBER];
static int id, emitting_face, listening_face, first_face, next_face, rot_servo, rot_dir, reconf_id;
static int distance;
static States state;

/* This function initialize all the global variables and starts the devices. */
static void reset(void)
{
  int i;

  /* The devices names. */
  char servos_name[] = "servo_00";
  char connectors_name[] = "connector_00";
  char emitters_name[] = "emitter_00";
  char receivers_name[] = "receiver_00";
  char *roombot_name = (char *) robot_get_name();

  /* The robot ID is taken from its name. */
  id = atoi(&roombot_name[7]);

  /* Initializing the servos. */
  for (i = 0; i < SERVO_NUMBER; i++) {
    servos[i] = robot_get_device(servos_name);
    servo_enable_position(servos[i], TIME_STEP);

    target_positions[i] = 0.0;
  
    servos_name[7]++;

    if ((i + 1) % 10 == 0) {
      servos_name[6]++;
      servos_name[7] = '0';
    }
  }

  /* Initializing the connectors. */
  for (i = 0; i < CONNECTOR_NUMBER; i++) {
    connectors[i] = robot_get_device(connectors_name);
    connector_enable_presence(connectors[i], TIME_STEP);

    connectors_name[11]++;

    if ((i + 1) % 10 == 0) {
      connectors_name[10]++;
      connectors_name[11] = '0';
    }
  }

  /* Initializing the emitters/receivers. */
  for (i = 0; i < FACES_NUMBER; i++) {
    emitters[i] = robot_get_device(emitters_name);
    receivers[i] = robot_get_device(receivers_name);
    receiver_enable(receivers[i], TIME_STEP);

    emitters_name[9]++;
    receivers_name[10]++;

    if ((i + 1) % 10 == 0) {
      emitters_name[8]++;
      emitters_name[9] = '0';
      receivers_name[9]++;
      receivers_name[10] = '0';
    }
  }
  
  distance = 0;
  
  switch (id) {
  case 0:
    state = DETACHING;
    emitting_face = 1;
    listening_face = 0;
    first_face = 3;
    next_face = 5;
    rot_servo = 2;
    rot_dir = -1;
    break;
  case 1:
    state = ROTATING_LOCK;
    emitting_face = 0;
    listening_face = 1;
    first_face = 5;
    next_face = 3;
    rot_servo = 2;
    rot_dir = 1;
    detach(first_face, ONLY);
    rotate(rot_servo, -90);
    distance++;
    break; 
  case 2:
    state = LISTENING;
    emitting_face = 5;
    listening_face = 4;
    first_face = 0;
    next_face = 2;
    rot_servo = 0;
    rot_dir = -1;
    break;
  case 3:
    state = LISTENING;
    emitting_face = 4;
    listening_face = 5;
    first_face = 2;
    next_face = 0;
    rot_servo = 0;
    rot_dir = 1;
    break;
  }
  reconf_id = 0;
  
  return;
}

/* This is the main loop function. */
static int run(int ms)
{
  unsigned char *buffer;
  States next_state;
  
  next_state = state;
  
  if (state == DETACHING) {
    detach(next_face, WITH);  
    next_state = LISTENING;
  } else if (state == LISTENING) {
    if (receiver_get_buffer_size(receivers[listening_face]) != 0) {
      if ((distance < 1) || (distance == 1 && id == 3)) {
        attach(first_face, WITHOUT);
        rotate(rot_servo, rot_dir * -90);
        next_state = ROTATING_LOCK;   
      } else if (distance == 1 && id == 1) {
        attach(first_face, WITHOUT);
        rotate(rot_servo, rot_dir * -180);
        next_state = ROTATING_LOCK;   
      } else {
        next_state = IDLE;
      }
      receiver_get_buffer(receivers[listening_face]);
      distance++;
    } 
  } else if (state == ROTATING_LOCK) {
    if (in_place(rot_servo) == 1) {
      attach(next_face, ONLY);
      detach(first_face, WITHOUT);
      if (distance == 2 && id == 1) {
        rotate(rot_servo, rot_dir * 180);
      }else{
        rotate(rot_servo, rot_dir * 90);
      }
      next_state = ROTATING;
    }
  } else if (state == ROTATING) {
    if (in_place(rot_servo) == 1) {
      buffer = (char *) emitter_get_buffer(emitters[emitting_face]);
      buffer[0] = 1;
      emitter_send(emitters[emitting_face], 1);
      if (distance == 2 && id == 3) {
        next_state = IDLE;
      } else {
        next_state = DETACHING;
      }
    }
  }

  state = next_state;

  return TIME_STEP;		/* run for TIME_STEP ms */
}

int main()
{
  robot_live(reset);
  robot_run(run);		/* this function never returns */

  return 0;
}

/*
 * This function simply returns 1 if there is at least one connectable device
 *  on the face_number and 0 if not.
 */
static int presence(int face_number, Behaviors behavior) 
{
  int i, special_index;

  switch (face_number) {
  case 0:
    special_index = 1;
    break;
  case 1:
    special_index = 6;
    break;
  case 2:
    special_index = 9;
    break;
  case 3:
    special_index = 15;
    break;
  case 4:
    special_index = 16;
    break;
  case 5:
    special_index = 20;
    break;
  }
  
  if (behavior != WITHOUT) {
    if (connector_get_presence(connectors[special_index]) == 1) {
      return 1;
    }
  } 
  
  if (behavior != ONLY) {
    for (i = 0; i < 4; i++) {
      if ((4 * face_number + i) != special_index) {
        if (connector_get_presence(connectors[4 * face_number + i]) == 1) {
          return 1;
        }
      }
    }
  }
  
  return 0;
}

/*
 * This function attaches the robot on face_number. Note that no test is
 * done to be shure that the face has really been attached.
 */
static void attach(int face_number, Behaviors behavior)
{
  int i, special_index;

  switch (face_number) {
  case 0:
    special_index = 1;
    break;
  case 1:
    special_index = 6;
    break;
  case 2:
    special_index = 9;
    break;
  case 3:
    special_index = 15;
    break;
  case 4:
    special_index = 16;
    break;
  case 5:
    special_index = 20;
    break;
  }
  
  if (behavior != WITHOUT) {
    connector_lock(connectors[special_index]);
  } 
  
  if (behavior != ONLY) {
    for (i = 0; i < 4; i++) {
      if ((4 * face_number + i) != special_index) {
        connector_lock(connectors[4 * face_number + i]);
      }
    }
  }

  return;
}

/* This function detaches the robot on face_number. */
static void detach(int face_number, Behaviors behavior)
{
  int i, special_index;

  switch (face_number) {
  case 0:
    special_index = 1;
    break;
  case 1:
    special_index = 6;
    break;
  case 2:
    special_index = 9;
    break;
  case 3:
    special_index = 15;
    break;
  case 4:
    special_index = 16;
    break;
  case 5:
    special_index = 20;
    break;
  }
  
  if (behavior != WITHOUT) {
    connector_unlock(connectors[special_index]);
  } 
  
  if (behavior != ONLY) {
    for (i = 0; i < 4; i++) {
      if ((4 * face_number + i) != special_index) {
        connector_unlock(connectors[4 * face_number + i]);
      }
    }
  }

  return;
}

/* This function rotates servo_number of degree_angle. */
static void rotate(int servo_number, float degree_angle)
{
  double rad_angle;

  /* The conversion in radians. */
  rad_angle = degree_angle * M_PI / 180;
  target_positions[servo_number] += rad_angle;

  if (target_positions[servo_number] > M_PI) {
    target_positions[servo_number] = M_PI;
  } else if (target_positions[servo_number] < -M_PI) {
    target_positions[servo_number] = -M_PI;
  }
  
  servo_set_position(servos[servo_number], target_positions[servo_number]);

  return;
}

/* 
 * This function returns 1 if servo_number has reached its target position
 * and 0 if not.
 */
static int in_place(int servo_number)
{
  if (fabs(servo_get_position(servos[servo_number]) - 
            target_positions[servo_number]) < TOL) {
    return 1;
  } else {
    return 0;
  }
}
