/*
 * File:         cube3_90_down_even.c
 * Date:         November 20th, 2006
 * Description:  A controller moving a group of Cube3 Roombot and climbing down a wall, this version
 *               is not the complete file (cube3_90_down.c), it is used as an example of how the 
 *               reconfiguration on the even side works.
 * Author:       Simon Blanchoud
 * Modifications:
 */

#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <device/robot.h>
#include <device/custom_robot.h>
#include <device/servo.h>
#include <device/connector.h>
#include <device/emitter.h>
#include <device/receiver.h>

#define TIME_STEP 64
#define SERVO_NUMBER 3
#define CONNECTOR_NUMBER 6
#define FACES_NUMBER 6
#define TOL 0.000001

/* An example of possible states*/
typedef enum
{ ROTATING, ATTACHING, DETACHING, IN_POSITION, RECONFIGURING_ODD, RECONFIGURING_EVEN, IDLE } States;

static void reset(void);
static int run(int);

/* The functions used to interact with the roombot. */
static int presence(int);
static void attach(int);
static void detach(int);
static void rotate(int, float);
static int in_place(int);

static DeviceTag servos[SERVO_NUMBER];
static DeviceTag connectors[CONNECTOR_NUMBER];
static DeviceTag emitters[FACES_NUMBER];
static DeviceTag receivers[FACES_NUMBER];

static double target_positions[SERVO_NUMBER];
static int id, distance, comm_id, servo_id, reconf_id;
static States state;

/* This function initialize all the global variables and starts the devices. */
static void reset(void)
{
  int i;

  /* The devices names. */
  char servos_name[] = "servo_00";
  char connectors_name[] = "connector_00";
  char emitters_name[] = "emitter_00";
  char receivers_name[] = "receiver_00";
  char *roombot_name = (char *) robot_get_name();

  /* The robot ID is taken from its name. */
  id = atoi(&roombot_name[7]);

  /* Initializing the servos. */
  for (i = 0; i < SERVO_NUMBER; i++) {
    servos[i] = robot_get_device(servos_name);
    servo_enable_position(servos[i], TIME_STEP);

    target_positions[i] = 0.0;
  
    servos_name[7]++;

    if ((i + 1) % 10 == 0) {
      servos_name[6]++;
      servos_name[7] = '0';
    }
  }

  /* Initializing the connectors. */
  for (i = 0; i < CONNECTOR_NUMBER; i++) {
    connectors[i] = robot_get_device(connectors_name);
    connector_enable_presence(connectors[i], TIME_STEP);

    connectors_name[11]++;

    if ((i + 1) % 10 == 0) {
      connectors_name[10]++;
      connectors_name[11] = '0';
    }
  }

  /* Initializing the emitters/receivers. */
  for (i = 0; i < FACES_NUMBER; i++) {
    emitters[i] = robot_get_device(emitters_name);
    receivers[i] = robot_get_device(receivers_name);
    receiver_enable(receivers[i], TIME_STEP);

    emitters_name[9]++;
    receivers_name[10]++;

    if ((i + 1) % 10 == 0) {
      emitters_name[8]++;
      emitters_name[9] = '0';
      receivers_name[9]++;
      receivers_name[10] = '0';
    }
  }

  /* We choose the initial state depending on the robot ID. */
  if (id == 0) {
    state = DETACHING;
    comm_id = 1;
    servo_id = 2;
  } else if (id == 1) {
    state = ROTATING;
    comm_id = 3;
    servo_id = 2;
  } else if (id == 2) {
    state = RECONFIGURING_EVEN;
    comm_id = 2;
    servo_id = 1;
  } else if (id == 3) {
    state = RECONFIGURING_EVEN;
    servo_id = 1;
  }
  attach(comm_id);

  distance = 0;
  reconf_id = 0;
  
  return;
}

/* This is the main loop function. */
static int run(int ms)
{
  unsigned char *buffer;
  States next_state;
  int next_reconf;

  next_state = state;
  
  /* Detache and ack the other robot. */
  if (state == DETACHING) {
    detach(servo_id);
    buffer = (unsigned char *) emitter_get_buffer(emitters[comm_id]);
    buffer[0] = 1;
    emitter_send(emitters[comm_id], 1);
    distance++;
    if (id == 0 && distance == 5) {
      next_state = RECONFIGURING_EVEN;
    } else if (distance < 6 - id) {
      next_state = ATTACHING;
    } else {
      next_state = IDLE;
    }
    /* If we have the ack that the other robot has detached, we turn. */
  } else if (state == ROTATING) {
    if (receiver_get_buffer_size(receivers[comm_id]) != 0) {
      rotate(servo_id, 180);
      next_state = IN_POSITION;
      receiver_get_buffer(receivers[comm_id]);
    }
    /* If we have the ack that the movement of the other is finished, we lock.  */
  } else if (state == ATTACHING) {
    if (receiver_get_buffer_size(receivers[comm_id]) != 0) {
      attach(servo_id);
      next_state = ROTATING;
      receiver_get_buffer(receivers[comm_id]);
    }
    /* When the movement is finished, we ack the other one. */
  } else if (state == IN_POSITION) {
    if (in_place(servo_id) == 1) {
      buffer = (char *) emitter_get_buffer(emitters[comm_id]);
      buffer[0] = 1;
      emitter_send(emitters[comm_id], 1);
      
      if (id == 1 && distance == 4) {
        next_state = RECONFIGURING_EVEN;
      } else if (id == 2 && distance == 0) {
        next_state = IDLE;
      } else {
        next_state = DETACHING;
      }
    }
  } else if (state == RECONFIGURING_EVEN) {
    next_reconf = 0;

    /*-------------The reconfiguration sequence of Cube 00-------------------*/
    if (id == 0) {
      if (reconf_id == 0) {
        if (receiver_get_buffer_size(receivers[1]) != 0) {
          receiver_get_buffer(receivers[1]);
          buffer = (char *) emitter_get_buffer(emitters[4]);
          buffer[0] = 1;
          emitter_send(emitters[4], 1);
          next_reconf = 1;
        }
      } else if (reconf_id == 1) {
        if (receiver_get_buffer_size(receivers[4]) != 0) {
          rotate(1, 180);
          receiver_get_buffer(receivers[4]);
          next_reconf = 1;
        }
      } else if (reconf_id == 2) {
        if (in_place(1) == 1) {
          buffer = (char *) emitter_get_buffer(emitters[4]);
          buffer[0] = 1;
          emitter_send(emitters[4], 1);
          next_reconf = 1;
        }
      } else if (reconf_id == 3) {
        if (receiver_get_buffer_size(receivers[4]) != 0) {
          receiver_get_buffer(receivers[4]);
          buffer = (char *) emitter_get_buffer(emitters[1]);
          buffer[0] = 1;
          emitter_send(emitters[1], 1);
          next_reconf = 1;
        }
      } else if (reconf_id == 4) {
        if (receiver_get_buffer_size(receivers[1]) != 0) {
          receiver_get_buffer(receivers[1]);
          buffer = (char *) emitter_get_buffer(emitters[4]);
          buffer[0] = 1;
          emitter_send(emitters[4], 1);
          next_reconf = 1;
        }
      } else if (reconf_id == 5) {
        if (receiver_get_buffer_size(receivers[4]) != 0) {
          receiver_get_buffer(receivers[4]);
          buffer = (char *) emitter_get_buffer(emitters[1]);
          buffer[0] = 1;
          emitter_send(emitters[1], 1);
          next_reconf = 1;
        }
      } else if (reconf_id == 6) {
        if (receiver_get_buffer_size(receivers[1]) != 0) {
          rotate(1, 180);
          receiver_get_buffer(receivers[1]);
          next_reconf = 1;
        }
      } else if (reconf_id == 7) {
        if (in_place(1) == 1) {
          buffer = (char *) emitter_get_buffer(emitters[1]);
          buffer[0] = 1;
          emitter_send(emitters[1], 1);
          next_reconf = 1;
        }
      } else if (reconf_id == 8) {
        if (receiver_get_buffer_size(receivers[3]) != 0) {
          comm_id = 4;
          servo_id = 1;
          next_state = DETACHING;
          receiver_get_buffer(receivers[3]);
        }
      }
    /*-------------The reconfiguration sequence of Cube 01-------------------*/
    } else if (id == 1) {
      if (reconf_id == 0) {
        if (receiver_get_buffer_size(receivers[3]) != 0) {
          detach(2);
          receiver_get_buffer(receivers[3]);
          buffer = (char *) emitter_get_buffer(emitters[3]);
          buffer[0] = 1;
          emitter_send(emitters[3], 1);
          next_reconf = 1;
        }
      } else if (reconf_id == 1) {
        if (receiver_get_buffer_size(receivers[3]) != 0) {
          attach(2);
          receiver_get_buffer(receivers[3]);
          buffer = (char *) emitter_get_buffer(emitters[4]);
          buffer[0] = 1;
          emitter_send(emitters[4], 1);
          next_reconf = 1;
        }
      } else if (reconf_id == 2) {
        if (receiver_get_buffer_size(receivers[4]) != 0) {
          receiver_get_buffer(receivers[4]);
          buffer = (char *) emitter_get_buffer(emitters[3]);
          buffer[0] = 1;
          emitter_send(emitters[3], 1);
          next_reconf = 1;
        }
      } else if (reconf_id == 3) {
        if (receiver_get_buffer_size(receivers[3]) != 0) {
          receiver_get_buffer(receivers[3]);
          buffer = (char *) emitter_get_buffer(emitters[4]);
          buffer[0] = 1;
          emitter_send(emitters[4], 1);
          next_reconf = 1;
        }
      } else if (reconf_id == 4) {
        if (receiver_get_buffer_size(receivers[4]) != 0) {
          detach(3);
          receiver_get_buffer(receivers[4]);
          buffer = (char *) emitter_get_buffer(emitters[4]);
          buffer[0] = 1;
          emitter_send(emitters[4], 1);
          next_reconf = 1;
        }
      } else if (reconf_id == 5) {
        if (receiver_get_buffer_size(receivers[4]) != 0) {
          attach(1);
          attach(3);
          receiver_get_buffer(receivers[4]);
          buffer = (char *) emitter_get_buffer(emitters[1]);
          buffer[0] = 1;
          emitter_send(emitters[1], 1);
          next_state = IDLE;
        }
      }
    /*-------------The reconfiguration sequence of Cube 02-------------------*/
    } else if (id == 2) {
      if (reconf_id == 0) {
        if (receiver_get_buffer_size(receivers[2]) != 0) {
          detach(1);
          receiver_get_buffer(receivers[2]);
          buffer = (char *) emitter_get_buffer(emitters[2]);
          buffer[0] = 1;
          emitter_send(emitters[2], 1);
          next_reconf = 1;
        }
      } else if (reconf_id == 1) {
        if (receiver_get_buffer_size(receivers[2]) != 0) {
          attach(1);
          receiver_get_buffer(receivers[2]);
          buffer = (char *) emitter_get_buffer(emitters[2]);
          buffer[0] = 1;
          emitter_send(emitters[2], 1);
          next_reconf = 1;
        }
      } else if (reconf_id == 2) {
        if (receiver_get_buffer_size(receivers[2]) != 0) {
          rotate(2, 180);
          receiver_get_buffer(receivers[2]);
          next_reconf = 1;
        }
      } else if (reconf_id == 3) {
        if (in_place(2) == 1) {
          buffer = (char *) emitter_get_buffer(emitters[2]);
          buffer[0] = 1;
          emitter_send(emitters[2], 1);
          next_reconf = 1;
        }
      } else if (reconf_id == 4) {
        if (receiver_get_buffer_size(receivers[3]) != 0) {
          rotate(2, 180);
          receiver_get_buffer(receivers[3]);
          next_reconf = 1;
        }
      } else if (reconf_id == 5) {
        if (in_place(2) == 1) {
          buffer = (char *) emitter_get_buffer(emitters[3]);
          buffer[0] = 1;
          emitter_send(emitters[3], 1);
          next_state = ROTATING;
        }
      }
    /*-------------The reconfiguration sequence of Cube 03-------------------*/
    } else if (id == 3) {
      if (reconf_id == 0) {
        if (receiver_get_buffer_size(receivers[2]) != 0) {
          rotate(2, 180);
          receiver_get_buffer(receivers[2]);
          next_reconf = 1;
        }
      } else if (reconf_id == 1) {
        if (in_place(2) == 1) {
          buffer = (char *) emitter_get_buffer(emitters[2]);
          buffer[0] = 1;
          emitter_send(emitters[2], 1);
          next_reconf = 1;
        }
      } else if (reconf_id == 2) {
        if (receiver_get_buffer_size(receivers[2]) != 0) {
          attach(1);
          receiver_get_buffer(receivers[2]);
          buffer = (char *) emitter_get_buffer(emitters[2]);
          buffer[0] = 1;
          emitter_send(emitters[2], 1);
          next_reconf = 1;
        }
      } else if (reconf_id == 3) {
        if (receiver_get_buffer_size(receivers[2]) != 0) {
          rotate(1, 90);
          receiver_get_buffer(receivers[2]);
          next_reconf = 1;
        }
      } else if (reconf_id == 4) {
        if (in_place(1) == 1) {
          buffer = (char *) emitter_get_buffer(emitters[1]);
          buffer[0] = 1;
          emitter_send(emitters[1], 1);
          next_reconf = 1;
        }
      } else if (reconf_id == 5) {
        if (receiver_get_buffer_size(receivers[1]) != 0) {
          rotate(2, -180);
          receiver_get_buffer(receivers[1]);
          next_reconf = 1;
        }
      } else if (reconf_id == 6) {
        if (in_place(2) == 1) {
          rotate(1, -90);
          next_reconf = 1;
        }
      } else if (reconf_id == 7) {
        if (in_place(1) == 1) {
          buffer = (char *) emitter_get_buffer(emitters[2]);
          buffer[0] = 1;
          emitter_send(emitters[2], 1);
          next_state = IDLE;
        }
      }
    }
    
    if (next_reconf == 1) {
      reconf_id++;
    }
  }
    
  state = next_state;

  return TIME_STEP;		/* run for TIME_STEP ms */
}

int main()
{
  robot_live(reset);
  robot_run(run);		/* this function never returns */

  return 0;
}

/*
 * This function simply returns 1 if there is a connectable device on the 
 * face_number and 0 if not.
 */
static int presence(int face_number) 
{
  if (connector_get_presence(connectors[face_number]) == 1) {
    return 1;
  } else {
    return 0;
  }
}

/*
 * This function attaches the robot on face_number. Note that no test is
 * done to be shure that the face has really been attached.
 */
static void attach(int face_number)
{
  connector_lock(connectors[face_number]);

  return;
}

/* This function detaches the robot on face_number. */
static void detach(int face_number)
{
  connector_unlock(connectors[face_number]);

  return;
}

/* This function rotates servo_number of degree_angle. */
static void rotate(int servo_number, float degree_angle)
{
  double rad_angle;

  /* The conversion in radians. */
  rad_angle = degree_angle * M_PI / 180;
  target_positions[servo_number] += rad_angle;

  servo_set_position(servos[servo_number], target_positions[servo_number]);

  return;
}

/* 
 * This function returns 1 if servo_number has reached its target position
 * and 0 if not.
 */
static int in_place(int servo_number)
{
  if (fabs(servo_get_position(servos[servo_number]) - 
            target_positions[servo_number]) < TOL) {
    return 1;
  } else {
    return 0;
  }
}
