/*
 * File:         cube2_90_right.c
 * Date:         November 29th, 2006
 * Description:  A controller moving 4 Cube2 Roombot on the right.
 * Author:       Simon Blanchoud
 * Modifications:
 */

#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <device/robot.h>
#include <device/custom_robot.h>
#include <device/servo.h>
#include <device/connector.h>
#include <device/emitter.h>
#include <device/receiver.h>

#define TIME_STEP 64
#define SERVO_NUMBER 3
#define CONNECTOR_NUMBER 24
#define FACES_NUMBER 6
#define TOL 0.000001

/* An example of possible states*/
typedef enum
{ ROTATING, DETACHING, ROTATING_LOCK, LISTENING, RECONFIGURING, IDLE} States;
typedef enum { WITH, WITHOUT, ONLY } Behaviors;

static void reset(void);
static int run(int);

/* The functions used to interact with the roombot. 
 *
 * The second parameter for the functions presence, attach and detach are 
 * related to the connector of the corresponding face which is placed on
 * the rotating servo. Its value can be:
 *      - WITH : Take also the rotating connector into account
 *      - WITHOUT : Do not take it into account
 *      - ONLY : Take only this connector into account
 */
static int presence(int, Behaviors);
static void attach(int, Behaviors);
static void detach(int, Behaviors);
static void rotate(int, float);
static int in_place(int);

static DeviceTag servos[SERVO_NUMBER];
static DeviceTag connectors[CONNECTOR_NUMBER];
static DeviceTag emitters[FACES_NUMBER];
static DeviceTag receivers[FACES_NUMBER];

static double target_positions[SERVO_NUMBER];
static int id, emitting_face, listening_face, first_face, next_face, rot_servo, rot_dir, reconf_id;
static int distance;
static States state;

/* This function initialize all the global variables and starts the devices. */
static void reset(void)
{
  int i;

  /* The devices names. */
  char servos_name[] = "servo_00";
  char connectors_name[] = "connector_00";
  char emitters_name[] = "emitter_00";
  char receivers_name[] = "receiver_00";
  char *roombot_name = (char *) robot_get_name();

  /* The robot ID is taken from its name. */
  id = atoi(&roombot_name[7]);

  /* Initializing the servos. */
  for (i = 0; i < SERVO_NUMBER; i++) {
    servos[i] = robot_get_device(servos_name);
    servo_enable_position(servos[i], TIME_STEP);

    target_positions[i] = 0.0;
  
    servos_name[7]++;

    if ((i + 1) % 10 == 0) {
      servos_name[6]++;
      servos_name[7] = '0';
    }
  }

  /* Initializing the connectors. */
  for (i = 0; i < CONNECTOR_NUMBER; i++) {
    connectors[i] = robot_get_device(connectors_name);
    connector_enable_presence(connectors[i], TIME_STEP);

    connectors_name[11]++;

    if ((i + 1) % 10 == 0) {
      connectors_name[10]++;
      connectors_name[11] = '0';
    }
  }

  /* Initializing the emitters/receivers. */
  for (i = 0; i < FACES_NUMBER; i++) {
    emitters[i] = robot_get_device(emitters_name);
    receivers[i] = robot_get_device(receivers_name);
    receiver_enable(receivers[i], TIME_STEP);

    emitters_name[9]++;
    receivers_name[10]++;

    if ((i + 1) % 10 == 0) {
      emitters_name[8]++;
      emitters_name[9] = '0';
      receivers_name[9]++;
      receivers_name[10] = '0';
    }
  }
  
  switch (id) {
  case 0:
    state = DETACHING;
    emitting_face = 1;
    listening_face = 0;
    first_face = 3;
    next_face = 5;
    rot_servo = 2;
    rot_dir = -1;
    break;
  case 1:
    state = ROTATING_LOCK;
    emitting_face = 0;
    listening_face = 1;
    first_face = 5;
    next_face = 3;
    rot_servo = 2;
    rot_dir = 1;
    detach(first_face, ONLY);
    rotate(rot_servo, -90);
    distance++;
    break; 
  case 2:
    state = LISTENING;
    emitting_face = 5;
    listening_face = 4;
    first_face = 0;
    next_face = 2;
    rot_servo = 0;
    rot_dir = -1;
    break;
  case 3:
    state = LISTENING;
    emitting_face = 4;
    listening_face = 5;
    first_face = 2;
    next_face = 0;
    rot_servo = 0;
    rot_dir = 1;
    break;
  }
  reconf_id = 0;
  
  return;
}

/* This is the main loop function. */
static int run(int ms)
{
  unsigned char *buffer;
  States next_state;
  int next_reconf;
  
  next_state = state;
  
  if (state == DETACHING) {
    detach(next_face, WITH);  
    next_state = LISTENING;
  } else if (state == LISTENING) {
    if (distance < 1) {
      if (receiver_get_buffer_size(receivers[listening_face]) != 0) {
        attach(first_face, WITHOUT);
        rotate(rot_servo, rot_dir * -90);
        next_state = ROTATING_LOCK;   
        receiver_get_buffer(receivers[listening_face]);
        distance++;
      } 
    } else { 
      next_state = RECONFIGURING;
    }
  } else if (state == ROTATING_LOCK) {
    if (in_place(rot_servo) == 1) {
      attach(next_face, ONLY);
      detach(first_face, WITHOUT);
      rotate(rot_servo, rot_dir * 90);
      next_state = ROTATING;
    }
  } else if (state == ROTATING) {
    if (in_place(rot_servo) == 1) {
      buffer = (char *) emitter_get_buffer(emitters[emitting_face]);
      buffer[0] = 1;
      emitter_send(emitters[emitting_face], 1);
      next_state = DETACHING;
    }
  } else if (state == RECONFIGURING) {
    next_reconf = 0;
    
    /*-------------The reconfiguration sequence of Cube 00-------------------*/
    if (id == 0) {
      if (reconf_id == 0) {
        buffer = (char *) emitter_get_buffer(emitters[emitting_face]);
        buffer[0] = 1;
        emitter_send(emitters[emitting_face], 1);
        next_reconf = 1;
      } else if (reconf_id == 1) {
        if (receiver_get_buffer_size(receivers[emitting_face]) != 0) {
          receiver_get_buffer(receivers[emitting_face]);
          buffer = (char *) emitter_get_buffer(emitters[0]);
          buffer[0] = 1;
          emitter_send(emitters[0], 1);
          next_reconf = 1;
        }
      } else if (reconf_id == 2) {
        if (receiver_get_buffer_size(receivers[4]) != 0) {
          detach(1, WITHOUT);
          detach(4, ONLY);
          detach(5, WITH);
          rotate(1, -90);
          receiver_get_buffer(receivers[4]);
          next_reconf = 1;
        }
      } else if (reconf_id == 3) {
        if (in_place(1) == 1) {
          attach(4, ONLY);
          buffer = (char *) emitter_get_buffer(emitters[4]);
          buffer[0] = 1;
          emitter_send(emitters[4], 1);
          next_reconf = 1;
        }
      } else if (reconf_id == 4) {
        if (receiver_get_buffer_size(receivers[4]) != 0) {
          detach(1, ONLY);
          rotate(1, 90);
          receiver_get_buffer(receivers[4]);
          next_reconf = 1;
        }
      } else if (reconf_id == 5) {
        if (in_place(0) == 1) {
          buffer = (char *) emitter_get_buffer(emitters[4]);
          buffer[0] = 1;
          emitter_send(emitters[4], 1);
          next_reconf = 1;
        }
      } else if (reconf_id == 6) {
        if (receiver_get_buffer_size(receivers[0]) != 0) {
          attach(0, WITH);
          attach(1, WITH);
          receiver_get_buffer(receivers[0]);
          next_reconf = 1;
        }
      } else if (reconf_id == 7) {
        if (receiver_get_buffer_size(receivers[4]) != 0) {
          detach(4, WITHOUT);
          rotate(1, 90);
          receiver_get_buffer(receivers[4]);
          next_reconf = 1;
        }
      } else if (reconf_id == 8) {
        if (in_place(1) == 1) {
          attach(1, WITH);
          buffer = (char *) emitter_get_buffer(emitters[1]);
          buffer[0] = 1;
          emitter_send(emitters[1], 1);
          next_reconf = 1;
        }
      } else if (reconf_id == 9) {
        if (receiver_get_buffer_size(receivers[1]) != 0) {
          detach(4, ONLY);
          rotate(1, -90);
          receiver_get_buffer(receivers[4]);
          next_reconf = 1;
        }
      } else if (reconf_id == 10) {
        if (in_place(1) == 1) {
          attach(4, WITH);
          buffer = (char *) emitter_get_buffer(emitters[1]);
          buffer[0] = 1;
          emitter_send(emitters[1], 1);
          next_state = IDLE;
        }
      }
    /*-------------The reconfiguration sequence of Cube 01-------------------*/
    } else if (id == 1) {
      if (reconf_id == 0) {
         if (receiver_get_buffer_size(receivers[1]) != 0) {
           receiver_get_buffer(receivers[1]);
           next_reconf = 1;
         }
      } else if (reconf_id == 1) {
        if (receiver_get_buffer_size(receivers[listening_face]) != 0) {
          attach(5, WITH);
          receiver_get_buffer(receivers[listening_face]);
          buffer = (char *) emitter_get_buffer(emitters[listening_face]);
          buffer[0] = 1;
          emitter_send(emitters[listening_face], 1);
          next_reconf = 1;
        }
      } else if (reconf_id == 2) {
         if (receiver_get_buffer_size(receivers[2]) != 0) {
          detach(5, WITH);
          receiver_get_buffer(receivers[2]);
          next_reconf = 1;
        }
      } else if (reconf_id == 3) {
         if (receiver_get_buffer_size(receivers[2]) != 0) {
          detach(2, WITH);
          receiver_get_buffer(receivers[2]);
          buffer = (char *) emitter_get_buffer(emitters[0]);
          buffer[0] = 1;
          emitter_send(emitters[0], 1);
          next_reconf = 1;
        }
      } else if (reconf_id == 4) {
        if (receiver_get_buffer_size(receivers[0]) != 0) {
          detach(0, ONLY);
          rotate(0, 180);
          receiver_get_buffer(receivers[0]);
          next_reconf = 1;
        }
      } else if (reconf_id == 5) {
        if (in_place(0) == 1) {
          detach(0, WITHOUT);
          attach(2, ONLY);
          rotate(0, -180);
          next_reconf = 1;
        }
      } else if (reconf_id == 6) {
        if (in_place(0) == 1) {
          attach(2, WITHOUT);
          buffer = (char *) emitter_get_buffer(emitters[2]);
          buffer[0] = 1;
          emitter_send(emitters[2], 1);
          next_reconf = 1;
        }
      } else if (reconf_id == 7) {
         if (receiver_get_buffer_size(receivers[1]) != 0) {
          receiver_get_buffer(receivers[1]);
          attach(1, WITH);
          receiver_get_buffer(receivers[1]);
          buffer = (char *) emitter_get_buffer(emitters[1]);
          buffer[0] = 1;
          emitter_send(emitters[1], 1);
          next_reconf = 1;
        }
      } else if (reconf_id == 8) {
         if (receiver_get_buffer_size(receivers[1]) != 0) {
          detach(2, WITHOUT);
          rotate(0, 180);
          receiver_get_buffer(receivers[1]);
          next_reconf = 1;
        }
      } else if (reconf_id == 9) {
        if (in_place(0) == 1) {
          attach(0, WITHOUT);
          detach(2, ONLY);
          rotate(0, -180);
          next_reconf = 1;
        }
      } else if (reconf_id == 10) {
        if (in_place(0) == 1) {
          buffer = (char *) emitter_get_buffer(emitters[0]);
          buffer[0] = 1;
          emitter_send(emitters[0], 1);
          next_state = IDLE;
        }
      }
    /*-------------The reconfiguration sequence of Cube 02-------------------*/
    }else if (id == 2) {
      if (reconf_id == 0) {
          if (receiver_get_buffer_size(receivers[5]) != 0) {
          detach(5, ONLY);
          rotate(2, -180);
          receiver_get_buffer(receivers[5]);
          next_reconf = 1;
        }
      } else if (reconf_id == 1) {
        if (in_place(2) == 1) {
          attach(3, ONLY);
          detach(5, WITHOUT);
          detach(4, WITH);
          rotate(2, 180);
          next_reconf = 1;
        }
      } else if (reconf_id == 2) {
        if (in_place(2) == 1) {
          attach(3, WITHOUT);
          buffer = (char *) emitter_get_buffer(emitters[3]);
          buffer[0] = 1;
          emitter_send(emitters[3], 1);
          next_reconf = 1;
        }
      } else if (reconf_id == 3) {
         if (receiver_get_buffer_size(receivers[3]) != 0) {
          attach(4, WITH);
          attach(0, WITH);
          receiver_get_buffer(receivers[3]);
          buffer = (char *) emitter_get_buffer(emitters[3]);
          buffer[0] = 1;
          emitter_send(emitters[3], 1);
          next_reconf = 1;
         }
      } else if (reconf_id == 4) {
        if (receiver_get_buffer_size(receivers[3]) != 0) {
          detach(3, ONLY);
          rotate(2, 180);
          receiver_get_buffer(receivers[3]);
          next_reconf = 1;
        }
      } else if (reconf_id == 5) {
        if (in_place(2) == 1) {
          attach(5, ONLY);
          detach(3, WITHOUT);
          rotate(2, -180);
          next_reconf = 1;
        }    
      } else if (reconf_id == 6) {
        if (in_place(2) == 1) {
          attach(5, WITHOUT);
          buffer = (char *) emitter_get_buffer(emitters[5]);
          buffer[0] = 1;
          emitter_send(emitters[5], 1);
          buffer = (char *) emitter_get_buffer(emitters[4]);
          buffer[0] = 1;
          emitter_send(emitters[4], 1);
          detach(0, ONLY);
          rotate(0, 90);
          next_reconf = 1;
        }
      } else if (reconf_id == 7) {
        if (in_place(0) == 1) {
          attach(2, ONLY);
          detach(0, WITHOUT);
          rotate(0, -90);
          next_reconf = 1;
        }
      } else if (reconf_id == 8) {
        if (in_place(0) == 1) {
          attach(2, WITHOUT);
          buffer = (char *) emitter_get_buffer(emitters[4]);
          buffer[0] = 1;
          emitter_send(emitters[4], 1);
          next_state = IDLE;
        }
      }
    /*-------------The reconfiguration sequence of Cube 03-------------------*/
    } else if (id == 3) {
      if (reconf_id == 0) {
        if (receiver_get_buffer_size(receivers[5]) != 0) {
          detach(1, ONLY);
          rotate(1, 180);
          receiver_get_buffer(receivers[5]);
          next_reconf = 1;
        }
      } else if (reconf_id == 1) {
        if (in_place(1) == 1) {
          attach(4, ONLY);
          detach(1, WITHOUT);
          rotate(1, -180);
          next_reconf = 1;
        }
      } else if (reconf_id == 2) {
        if (in_place(1) == 1) {
          attach(4, WITH);
          attach(1, WITH);
          buffer = (char *) emitter_get_buffer(emitters[5]);
          buffer[0] = 1;
          emitter_send(emitters[5], 1);
          next_reconf = 1;
        }
      } else if (reconf_id == 3) {
         if (receiver_get_buffer_size(receivers[1]) != 0) {
          receiver_get_buffer(receivers[1]);
          buffer = (char *) emitter_get_buffer(emitters[4]);
          buffer[0] = 1;
          emitter_send(emitters[4], 1);
          next_reconf = 1;
        }
      } else if (reconf_id == 4) {
        if (receiver_get_buffer_size(receivers[1]) != 0) {
          attach(3, WITHOUT);
          receiver_get_buffer(receivers[1]);
          next_reconf = 1;
        }
      } else if (reconf_id == 5) {
         if (receiver_get_buffer_size(receivers[3]) != 0) {
          detach(4, WITH);
          receiver_get_buffer(receivers[3]);
          buffer = (char *) emitter_get_buffer(emitters[3]);
          buffer[0] = 1;
          emitter_send(emitters[3], 1);
          next_reconf = 1;
        }
      } else if (reconf_id == 6) {
         if (receiver_get_buffer_size(receivers[3]) != 0) {
          detach(3, ONLY);
          rotate(2, 180);
          receiver_get_buffer(receivers[3]);
          next_reconf = 1;
        }
      } else if (reconf_id == 7) {
        if (in_place(2) == 1) {
          attach(5, ONLY);
          detach(3, WITHOUT);
          rotate(2, -180);
          next_reconf = 1;
        }
      } else if (reconf_id == 8) {
        if (in_place(2) == 1) {
          attach(5, WITHOUT);
          attach(4, WITH);
          next_state = IDLE;
        }
      }
    }

    if (next_reconf == 1) {
      reconf_id++;
    }
  }

  state = next_state;

  return TIME_STEP;		/* run for TIME_STEP ms */
}

int main()
{
  robot_live(reset);
  robot_run(run);		/* this function never returns */

  return 0;
}

/*
 * This function simply returns 1 if there is at least one connectable device
 *  on the face_number and 0 if not.
 */
static int presence(int face_number, Behaviors behavior) 
{
  int i, special_index;

  switch (face_number) {
  case 0:
    special_index = 1;
    break;
  case 1:
    special_index = 6;
    break;
  case 2:
    special_index = 9;
    break;
  case 3:
    special_index = 15;
    break;
  case 4:
    special_index = 16;
    break;
  case 5:
    special_index = 20;
    break;
  }
  
  if (behavior != WITHOUT) {
    if (connector_get_presence(connectors[special_index]) == 1) {
      return 1;
    }
  } 
  
  if (behavior != ONLY) {
    for (i = 0; i < 4; i++) {
      if ((4 * face_number + i) != special_index) {
        if (connector_get_presence(connectors[4 * face_number + i]) == 1) {
          return 1;
        }
      }
    }
  }
  
  return 0;
}

/*
 * This function attaches the robot on face_number. Note that no test is
 * done to be shure that the face has really been attached.
 */
static void attach(int face_number, Behaviors behavior)
{
  int i, special_index;

  switch (face_number) {
  case 0:
    special_index = 1;
    break;
  case 1:
    special_index = 6;
    break;
  case 2:
    special_index = 9;
    break;
  case 3:
    special_index = 15;
    break;
  case 4:
    special_index = 16;
    break;
  case 5:
    special_index = 20;
    break;
  }
  
  if (behavior != WITHOUT) {
    connector_lock(connectors[special_index]);
  } 
  
  if (behavior != ONLY) {
    for (i = 0; i < 4; i++) {
      if ((4 * face_number + i) != special_index) {
        connector_lock(connectors[4 * face_number + i]);
      }
    }
  }

  return;
}

/* This function detaches the robot on face_number. */
static void detach(int face_number, Behaviors behavior)
{
  int i, special_index;

  switch (face_number) {
  case 0:
    special_index = 1;
    break;
  case 1:
    special_index = 6;
    break;
  case 2:
    special_index = 9;
    break;
  case 3:
    special_index = 15;
    break;
  case 4:
    special_index = 16;
    break;
  case 5:
    special_index = 20;
    break;
  }
  
  if (behavior != WITHOUT) {
    connector_unlock(connectors[special_index]);
  } 
  
  if (behavior != ONLY) {
    for (i = 0; i < 4; i++) {
      if ((4 * face_number + i) != special_index) {
        connector_unlock(connectors[4 * face_number + i]);
      }
    }
  }

  return;
}

/* This function rotates servo_number of degree_angle. */
static void rotate(int servo_number, float degree_angle)
{
  double rad_angle;

  /* The conversion in radians. */
  rad_angle = degree_angle * M_PI / 180;
  target_positions[servo_number] += rad_angle;

  if (target_positions[servo_number] > M_PI) {
    target_positions[servo_number] = M_PI;
  } else if (target_positions[servo_number] < -M_PI) {
    target_positions[servo_number] = -M_PI;
  }
  
  servo_set_position(servos[servo_number], target_positions[servo_number]);

  return;
}

/* 
 * This function returns 1 if servo_number has reached its target position
 * and 0 if not.
 */
static int in_place(int servo_number)
{
  if (fabs(servo_get_position(servos[servo_number]) - 
            target_positions[servo_number]) < TOL) {
    return 1;
  } else {
    return 0;
  }
}
