/*
 * File:         cube1_straight.c
 * Date:         November 15th, 2006
 * Description:  A controller moving two Cube1 Roombot straight.
 * Author:       Simon Blanchoud
 * Modifications:
 */

#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <device/robot.h>
#include <device/custom_robot.h>
#include <device/servo.h>
#include <device/connector.h>
#include <device/emitter.h>
#include <device/receiver.h>

#define TIME_STEP 64
#define SERVO_NUMBER 1
#define CONNECTOR_NUMBER 6
#define FACES_NUMBER 6
#define TOL 0.001

typedef enum
{ ROTATING, ATTACHING, DETACHING, IN_POSITION, IDLE } States;

static void reset(void);
static int run(int);

static int presence(int);
static void attach(int);
static void detach(int);
static void rotate(int, float);
static int in_place(int);

static DeviceTag servos[SERVO_NUMBER];
static DeviceTag connectors[CONNECTOR_NUMBER];
static DeviceTag emitters[FACES_NUMBER];
static DeviceTag receivers[FACES_NUMBER];

static double target_positions[SERVO_NUMBER];
static int id, distance, comm_id;
static States state;

/* This function initialize all the global variables and starts the devices. */
static void reset(void)
{
  int i;

  /* The devices names. */
  char servos_name[] = "servo_00";
  char connectors_name[] = "connector_00";
  char emitters_name[] = "emitter_00";
  char receivers_name[] = "receiver_00";
  char *roombot_name = (char *) robot_get_name();

  /* The robot ID is taken from its name. */
  id = atoi(&roombot_name[7]);

  /* Initializing the servos. */
  for (i = 0; i < SERVO_NUMBER; i++) {
    servos[i] = robot_get_device(servos_name);
    servo_enable_position(servos[i], TIME_STEP);

    target_positions[i] = 0.0;
  
    servos_name[7]++;

    if ((i + 1) % 10 == 0) {
      servos_name[6]++;
      servos_name[7] = '0';
    }
  }

  /* Initializing the connectors. */
  for (i = 0; i < CONNECTOR_NUMBER; i++) {
    connectors[i] = robot_get_device(connectors_name);
    connector_enable_presence(connectors[i], TIME_STEP);

    connectors_name[11]++;

    if ((i + 1) % 10 == 0) {
      connectors_name[10]++;
      connectors_name[11] = '0';
    }
  }

  /* Initializing the emitters/receivers. */
  for (i = 0; i < FACES_NUMBER; i++) {
    emitters[i] = robot_get_device(emitters_name);
    receivers[i] = robot_get_device(receivers_name);
    receiver_enable(receivers[i], TIME_STEP);

    emitters_name[9]++;
    receivers_name[10]++;

    if ((i + 1) % 10 == 0) {
      emitters_name[8]++;
      emitters_name[9] = '0';
      receivers_name[9]++;
      receivers_name[10] = '0';
    }
  }

  /* We choose the initial state depending on the robot ID. */
  if (id == 0) {
    state = DETACHING;
    comm_id = 1;
  } else {
    state = ROTATING;
    comm_id = 3;
  }

  distance = 0;
  
  return;
}

/* This is the main loop function. */
static int run(int ms)
{
  unsigned char *buffer;
  States next_state;

  next_state = state;
  
  /* Detache and ack the other robot. */
  if (state == DETACHING) {
    detach(5);
    buffer = (unsigned char *) emitter_get_buffer(emitters[comm_id]);
    buffer[0] = 1;
    emitter_send(emitters[comm_id], 1);
    distance++;
    if (distance < 6 - id) {
      next_state = ATTACHING;
    } else {
      next_state = IDLE;
    }
    /* If we have the ack that the other robot has detached, we turn. */
  } else if (state == ROTATING) {
    if (receiver_get_buffer_size(receivers[comm_id]) != 0) {
      rotate(0, 180);
      next_state = IN_POSITION;
      receiver_get_buffer(receivers[comm_id]);
    }
    /* If we have the ack that the movement of the other is finished, we lock.  */
  } else if (state == ATTACHING) {
    if (receiver_get_buffer_size(receivers[comm_id]) != 0) {
      attach(5);
      next_state = ROTATING;
      receiver_get_buffer(receivers[comm_id]);
    }
    /* When the movement is finished, we ack the other one. */
  } else if (state == IN_POSITION) {
    if (in_place(0) == 1) {
      buffer = (char *) emitter_get_buffer(emitters[comm_id]);
      buffer[0] = 1;
      emitter_send(emitters[comm_id], 1);
      next_state = DETACHING;
    }
  }
    
  state = next_state;

  return TIME_STEP;		/* run for TIME_STEP ms */
}

int main()
{
  robot_live(reset);
  robot_run(run);		/* this function never returns */

  return 0;
}

/*
 * This function simply returns 1 if there is a connectable device on the 
 * face_number and 0 if not.
 */
static int presence(int face_number) 
{
  if (connector_get_presence(connectors[face_number]) == 1) {
    return 1;
  } else {
    return 0;
  }
}

/*
 * This function attaches the robot on face_number. Note that no test is
 * done to be shure that the face has really been attached.
 */
static void attach(int face_number)
{
  connector_lock(connectors[face_number]);

  return;
}

/* This function detaches the robot on face_number. */
static void detach(int face_number)
{
  connector_unlock(connectors[face_number]);

  return;
}

/* This function rotates servo_number of degree_angle. */
static void rotate(int servo_number, float degree_angle)
{
  double rad_angle;

  /* The conversion in radians. */
  rad_angle = degree_angle * M_PI / 180;
  target_positions[servo_number] += rad_angle;

  servo_set_position(servos[servo_number], target_positions[servo_number]);

  return;
}

/* 
 * This function returns 1 if servo_number has reached its target position
 * and 0 if not.
 */
static int in_place(int servo_number)
{
  if (fabs(servo_get_position(servos[servo_number]) - 
            target_positions[servo_number]) < TOL) {
    return 1;
  } else {
    return 0;
  }
}
