/*
 * File:         cube2.c
 * Date:         January 9th, 2007
 * Description:  A controller handeling the Cube2 Roombot.
 * Author:       Simon Blanchoud
 * Modifications:
 */

#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <device/robot.h>
#include <device/custom_robot.h>
#include <device/servo.h>
#include <device/connector.h>
#include <device/emitter.h>
#include <device/receiver.h>

#define TIME_STEP 64
#define SERVO_NUMBER 3
#define CONNECTOR_NUMBER 24
#define FACES_NUMBER 6
#define TOL 0.001

typedef enum {WITH, WITHOUT, ONLY, ERROR} Behaviors;

static void reset(void);
static int run(int);

/*
 * The functions used to interact with the roombot. 
 *
 * The second parameter for the functions presence, attach and detach are 
 * related to the connector of the corresponding face which is placed on
 * the rotating servo. Its value can be:
 *      - WITH : Take also the rotating connector into account
 *      - WITHOUT : Do not take it into account
 *      - ONLY : Take only this connector into account
 *      - ERROR : Should never be used, appears when there is a translation 
 *                problem, nothing will be done in this case
 */
static void attach(int, Behaviors);
static void detach(int, Behaviors);
static void rotate(int, int);
static void in_place(int);
static Behaviors translate(int);

static DeviceTag servos[SERVO_NUMBER];
static DeviceTag connectors[CONNECTOR_NUMBER];
static DeviceTag emitter;
static DeviceTag receiver;

static double target_positions[SERVO_NUMBER];
static int is_rotating[SERVO_NUMBER];
static int id;

/* This function initialize all the global variables and starts the devices. */
static void reset(void)
{
  int i;

  /* The devices names. */
  char servos_name[] = "servo_00";
  char connectors_name[] = "connector_00";
  char *roombot_name = (char *) robot_get_name();

  /* The robot ID is taken from its name. */
  id = atoi(&roombot_name[7]);

  if (id == 0) {
    robot_console_printf("Fatal Error: the ID 0 is reserved for the supervisor, please change it\n");
    exit(0);
  }

  /* Initializing the servos. */
  for (i = 0; i < SERVO_NUMBER; i++) {
    servos[i] = robot_get_device(servos_name);
    servo_enable_position(servos[i], TIME_STEP);

    target_positions[i] = 0.0;
    is_rotating[i] = 0;
  
    servos_name[7]++;

    if ((i + 1) % 10 == 0) {
      servos_name[6]++;
      servos_name[7] = '0';
    }
  }

  /* Initializing the connectors. */
  for (i = 0; i < CONNECTOR_NUMBER; i++) {
    connectors[i] = robot_get_device(connectors_name);
    connector_enable_presence(connectors[i], TIME_STEP);

    connectors_name[11]++;

    if ((i + 1) % 10 == 0) {
      connectors_name[10]++;
      connectors_name[11] = '0';
    }
  }

  /* Initializing the emitter and the receiver. */
  emitter = robot_get_device("emitter");
  emitter_set_channel(emitter, 0);
  
  receiver = robot_get_device("receiver");
  receiver_set_channel(receiver, id);
  receiver_enable(receiver, TIME_STEP); 

  return;
}

/* This is the main loop function. */
static int run(int ms)
{
  int i, value;
  
  if (receiver_get_queue_length(receiver) > 0) {
      const unsigned char *data = receiver_get_data(receiver);

      switch (data[0]) {
      case 'A' :
        attach(data[2], translate(data[3]));
        break;
      case 'D' :
        detach(data[2], translate(data[3]));
        break;
      case 'R' :
        value = ((int) data[3] << 8) + (int) data[4];
        if (value > 32767) {
          value -= 65536;
        }
        
        rotate(data[2], value);
        break;
      default :
        robot_console_printf("Error: unknown command %c\n", data[0]);
        break;
      }

      receiver_next_packet(receiver);
  }

  for (i = 0; i < SERVO_NUMBER; i++) {
    if (is_rotating[i] == 1) {
      in_place(i);
    }
  } 
  
  return TIME_STEP;		/* run for TIME_STEP ms */
}

int main()
{
  robot_live(reset);
  robot_run(run);		/* this function never returns */

  return 0;
}

/*
 * This function is used to translate between the integer used in the sequence
 * files and the Behaviors type used in this file.
 */
static Behaviors translate(int received_behavior)
{
  switch (received_behavior) {
  case 0 :
    return WITH;
    break;
  case 1 :
    return WITHOUT;
    break;
  case 2 :
    return ONLY;
    break;
  default :
    robot_console_printf("Error: unknown behavior code %d\n", received_behavior);
    return ERROR;
    break;
  }
}

/*
 * This function attaches the robot on face_number. Note that no test is
 * done to be shure that the face has really been attached.
 */
static void attach(int face_number, Behaviors behavior)
{
  unsigned char buffer[2];
  int i, special_index;

  if (behavior == ERROR) {
    return;
  }
  
  switch (face_number) {
  case 0:
    special_index = 1;
    break;
  case 1:
    special_index = 6;
    break;
  case 2:
    special_index = 9;
    break;
  case 3:
    special_index = 15;
    break;
  case 4:
    special_index = 16;
    break;
  case 5:
    special_index = 20;
    break;
  }
  
  if (behavior != WITHOUT) {
    connector_lock(connectors[special_index]);
  } 
  
  if (behavior != ONLY) {
    for (i = 0; i < 4; i++) {
      if ((4 * face_number + i) != special_index) {
        connector_lock(connectors[4 * face_number + i]);
      }
    }
  }

  buffer[0] = 'A';
  buffer[1] = id;
  
  emitter_send_packet(emitter, (const void *) buffer, 2);
  
  return;
}

/* This function detaches the robot on face_number. */
static void detach(int face_number, Behaviors behavior)
{
  unsigned char buffer[2];
  int i, special_index;

  if (behavior == ERROR) {
    return;
  }

  switch (face_number) {
  case 0:
    special_index = 1;
    break;
  case 1:
    special_index = 6;
    break;
  case 2:
    special_index = 9;
    break;
  case 3:
    special_index = 15;
    break;
  case 4:
    special_index = 16;
    break;
  case 5:
    special_index = 20;
    break;
  }
  
  if (behavior != WITHOUT) {
    connector_unlock(connectors[special_index]);
  } 
  
  if (behavior != ONLY) {
    for (i = 0; i < 4; i++) {
      if ((4 * face_number + i) != special_index) {
        connector_unlock(connectors[4 * face_number + i]);
      }
    }
  }
  
  buffer[0] = 'D';
  buffer[1] = id;
  
  emitter_send_packet(emitter, (const void *) buffer, 2);
  
  return;
}

/* This function rotates servo_number of degree_angle. */
static void rotate(int servo_number, int degree_angle)
{
  double rad_angle;

  /* The conversion in radians. */
  rad_angle = (double) degree_angle * M_PI / 180;
  target_positions[servo_number] += rad_angle;

  if (target_positions[servo_number] > M_PI) {
    target_positions[servo_number] = M_PI;
  } else if (target_positions[servo_number] < -M_PI) {
    target_positions[servo_number] = -M_PI;
  }
  
  servo_set_position(servos[servo_number], target_positions[servo_number]);
  is_rotating[servo_number] = 1;

  return;
}

/* 
 * This function returns 1 if servo_number has reached its target position
 * and 0 if not.
 */
static void in_place(int servo_number)
{
  unsigned char buffer[2];
  
  buffer[0] = 'R';
  buffer[1] = id;
  
  if (fabs(servo_get_position(servos[servo_number]) - 
            target_positions[servo_number]) < TOL) {
  
    emitter_send_packet(emitter, (const void *) buffer, 2);
    is_rotating[servo_number] = 0;
  }
  
  return;
}

