/*
 * File:         cube3.c
 * Date:         January 9th, 2007
 * Description:  A controller handeling the Cube3 Roombot.
 * Author:       Simon Blanchoud
 * Modifications:
 */

#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <device/robot.h>
#include <device/custom_robot.h>
#include <device/servo.h>
#include <device/connector.h>
#include <device/emitter.h>
#include <device/receiver.h>

#define TIME_STEP 64
#define SERVO_NUMBER 3
#define CONNECTOR_NUMBER 6
#define FACES_NUMBER 6
#define TOL 0.001

static void reset(void);
static int run(int);

/* The functions used to interact with the roombot. */
static void attach(int);
static void detach(int);
static void rotate(int, int);
static void in_place(int);

static DeviceTag servos[SERVO_NUMBER];
static DeviceTag connectors[CONNECTOR_NUMBER];
static DeviceTag emitter;
static DeviceTag receiver;

static double target_positions[SERVO_NUMBER];
static int is_rotating[SERVO_NUMBER];
static int id;

/* This function initialize all the global variables and starts the devices. */
static void reset(void)
{
  int i;

  /* The devices names. */
  char servos_name[] = "servo_00";
  char connectors_name[] = "connector_00";
  char *roombot_name = (char *) robot_get_name();

  /* The robot ID is taken from its name. */
  id = atoi(&roombot_name[7]);

  if (id == 0) {
    robot_console_printf("Fatal Error: the ID 0 is reserved for the supervisor, please change it\n");
    exit(0);
  }

  /* Initializing the servos. */
  for (i = 0; i < SERVO_NUMBER; i++) {
    servos[i] = robot_get_device(servos_name);
    servo_enable_position(servos[i], TIME_STEP);

    target_positions[i] = 0.0;
    is_rotating[i] = 0;
  
    servos_name[7]++;

    if ((i + 1) % 10 == 0) {
      servos_name[6]++;
      servos_name[7] = '0';
    }
  }

  /* Initializing the connectors. */
  for (i = 0; i < CONNECTOR_NUMBER; i++) {
    connectors[i] = robot_get_device(connectors_name);
    connector_enable_presence(connectors[i], TIME_STEP);

    connectors_name[11]++;

    if ((i + 1) % 10 == 0) {
      connectors_name[10]++;
      connectors_name[11] = '0';
    }
  }
  
  /* Initializing the emitter and the receiver. */
  emitter = robot_get_device("emitter");
  emitter_set_channel(emitter, 0);
  
  receiver = robot_get_device("receiver");
  receiver_set_channel(receiver, id);
  receiver_enable(receiver, TIME_STEP);
  
  return;
}

/* This is the main loop function. */
static int run(int ms)
{
  int i, value;
  
  if (receiver_get_queue_length(receiver) > 0) {
      const unsigned char *data = receiver_get_data(receiver);

      switch (data[0]) {
      case 'A' :
        attach(data[2]);
        break;
      case 'D' :
        detach(data[2]);
        break;
      case 'R' :
        value = ((int) data[3] << 8) + (int) data[4];
        if (value > 32767) {
          value -= 65536;
        }
        
        rotate(data[2], value);
        break;
      default :
        robot_console_printf("Error: unknown command %c\n", data[0]);
        break;
      }

      receiver_next_packet(receiver);
  }

  for (i = 0; i < SERVO_NUMBER; i++) {
    if (is_rotating[i] == 1) {
      in_place(i);
    }
  }

  return TIME_STEP;		/* run for TIME_STEP ms */
}

int main()
{
  robot_live(reset);
  robot_run(run);		/* this function never returns */

  return 0;
}

/*
 * This function attaches the robot on face_number. Note that no test is
 * done to be shure that the face has really been attached.
 */
static void attach(int face_number)
{
  unsigned char buffer[2];
  
  buffer[0] = 'A';
  buffer[1] = id;
  
  connector_lock(connectors[face_number]);
  emitter_send_packet(emitter, (const void *) buffer, 2);

  return;
}

/* This function detaches the robot on face_number. */
static void detach(int face_number)
{
  unsigned char buffer[2];
  
  buffer[0] = 'D';
  buffer[1] = id;
  
  connector_unlock(connectors[face_number]);
  emitter_send_packet(emitter, (const void *) buffer, 2);

  return;
}

/* This function rotates servo_number of degree_angle. */
static void rotate(int servo_number, int degree_angle)
{
  double rad_angle;

  /* The conversion in radians. */
  rad_angle = (double) degree_angle * M_PI / 180;
  target_positions[servo_number] += rad_angle;

  servo_set_position(servos[servo_number], target_positions[servo_number]);
  is_rotating[servo_number] = 1;

  return;
}

/* 
 * This function returns 1 if servo_number has reached its target position
 * and 0 if not.
 */
static void in_place(int servo_number)
{
  unsigned char buffer[2];
  
  buffer[0] = 'R';
  buffer[1] = id;
  
  if (fabs(servo_get_position(servos[servo_number]) - 
            target_positions[servo_number]) < TOL) {
  
    emitter_send_packet(emitter, (const void *) buffer, 2);
    is_rotating[servo_number] = 0;
  }

  return;
}

