/*
 * File:         supervisor.c
 * Date:         January 19th, 2007
 * Description:  Supervisor for the Roombots modules.
 * Author:       Simon Blanchoud
 */

/*
 * This supervisor is used to synchronize the movements of all the roombots.
 * There are two ways of using it:
 *  - By creating a sequence file, which name is passed using the 
 *    controllerArgs. The specification of the protocol used in these files
 *    is specificated in sequence_protocol.txt. Note that various files can
 *    be specified in controllerArgs, in this cas the supervisor will play them
 *    one after the other.
 *  - By writing explicitly the actions you want to perform in the run function.
 *    To simplify this you can reuse the functions created to interact with the
 *    modules.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <device/robot.h>
#include <device/supervisor.h>
#include <device/emitter.h>
#include <device/receiver.h>

#define TIME_STEP 64

/* The size of the buffer used to read the POSITIONS_FILE. */
#define READ_BUFFER_SIZE 1024
/*
 * The size used to allocate the memory of the sequence. This is not its
 * maximal size, when this size is reached it is automatically incremented
 * by itself.
 */
#define SEQUENCE_SIZE 32

/* The structure used to store the commands of a sequence. */
typedef struct {
  char tag;
  int roombot_id;
  int device_id;
  
  /* The value in degree of the rotation. */
  int value;
  
  /* 
   * The behavior of the attachment/detachment, see the sequence_protocol.txt 
   * file for more details.
   */
  int behavior;

  /* The string we want to print on the log. */
  char *string;

} Command;

/* The functions used by Webots. */
static void reset(void);
static int run(int);

/* The functions used to play a sequence from a file. */
static int load_sequence(char *);
static void execute_sequence(void);
static void clean_sequence(void);

/* The functions used to interact with the modules. */

/* The functions used to start an action on a module. */
static void attach(int, int, int);
static void detach(int, int, int);
static void rotate(int, int, int);
/* The functions used to check if an action has been performed. */
static void is_attached(int);
static void is_detached(int);
static void has_rotated(int);
/* This function places a new module on the initial position. */
static void place_new(int);
  
static DeviceTag emitter, receiver;
static Command *actions;
/* The size of the actions array. */
static int sequence_size;

/* The number of the last module we placed. */
static int placed_modules;

/* The initialization function. */
static void reset(void)
{
  /* We first retreive the devices. */
  emitter = robot_get_device("emitter");
  receiver = robot_get_device("receiver");

  /* 
   * The supervisor will send its commands on the channel corresponding to the 
   * id of the module with which it wants to interact, so it will modify it 
   * directly when it will know this id. But all the modules will answer on the
   * same channel (0) so that we will not loose any packet because we were not
   * listening on the right channel.
   * Please note that as the channel used by the modules is based on their id 
   * and as the supervisor is using the channel 0, you should not create any
   * module with the id 0.
   */
  receiver_enable(receiver, TIME_STEP);
  receiver_set_channel(receiver, 0);

  sequence_size = 0;
  placed_modules = 0;

  /* This cleans the log. */
  robot_console_printf("\f");
  
  return;
}

/* 
 * This function is called at every time step. You can use it to interact 
 * directly with the modules either by creating a finite state machine or
 * by writing a simple sequence of interaction functions which will be 
 * repeated endlessly.
 */
static int run(int ms)
{

  return TIME_STEP;
}

/*
 * The main function, called first and which receives the arguments of
 * controllerArgs.
 */
int main(int argc, char *argv[])
{
  int loaded, read_files;
  
  /* Initialization. */
  robot_live(reset);
  
  /* If there are arguments, this means that we have sequences to load. */
  read_files = 1;
  while (argc > read_files) {
    loaded = load_sequence(argv[read_files]);
    read_files++;
    
    if (loaded == 1) {
      robot_console_printf("+++ running %s +++\n", argv[read_files - 1]);
      execute_sequence();
      robot_console_printf("--- %s end ---\n", argv[read_files - 1]);
    }
    
    clean_sequence();
  }
  
  robot_run(run);             /* never returns */

  return 0;
}

/*
 * This function loads the actions contained in the file located at path and 
 * store them in the actions array, adapting its size as needed, and sets 
 * sequence_size as needed. 
 * It returns 1 in case of success and 0 in the other case.
 */
static int load_sequence(char *path)
{
  FILE *fp;
  char buffer[READ_BUFFER_SIZE];
  char *next_token;
  int actions_read, actions_skipped, skipping;
  int string_length;
  Command *new_actions;
  Command temp_command;

  fp = fopen(path, "r");

  actions_read = 0;
  actions_skipped = 0;

  /* We first of all start with a size of SEQUENCE_SIZE for the array. */
  sequence_size = SEQUENCE_SIZE;
  actions = (Command *) malloc(SEQUENCE_SIZE * sizeof(Command));

  /* If this is not possible we return with an error value. */
  if (actions == NULL) {
    robot_console_printf("Error: Unable to allocate memory for the sequence\n");

    return 0;
  }
  
  /* The file was not opened. */
  if (fp == NULL) {
    robot_console_printf("Error: Cannot open the file %s\n", path);

    return 0;
  }

  /* 
   * We read the first line, which has a maximum length of READ_BUFFER_SIZE,
   * including the \0 end character.
    */
  fgets(buffer, READ_BUFFER_SIZE - 1, fp);

  /* While we have not reached the end of the file we continue. */
  while (feof(fp) == 0) {

    /*
     * If we have more than 1 char on the line and if it is not a commentary we
     * can start storing this command. We need to have more than 1, and not 0,
     * because the \n char is also counted and so a blank line would be valid.
     */
    if (strlen(buffer) > 1 && strncmp(buffer, "#", 1) != 0) {
      skipping = 0;
      
      /* Here we set the default values of an action. */
      temp_command.tag = -1;
      temp_command.string = NULL;
      temp_command.roombot_id = -1;
      temp_command.device_id = -1;
      temp_command.value = 0;
      temp_command.behavior = 0;
      
      /* 
       * The case of the 'O' command is special as we want to directly print 
       * in the log what is written in the file, without seperating on the 
       * blank spaces. So we need to handle it before we split the buffer.
       */
      if (strncmp(buffer, "O", 1) == 0) {

        /* 
         * This is the size of the string (without the tag, the space and the
         * new line characters). 
         */
        string_length = strlen(buffer) - 3;
        if (string_length > 0) {
          temp_command.tag = buffer[0];
          
          /* 
           * We allocate enough space to put the entire string. If this is not
           * possible, we skip this command. 
           */
          temp_command.string = (char *) malloc((string_length + 1) * sizeof(char));
          if (temp_command.string == NULL) {
            robot_console_printf("Error: Unable to allocate memory for the string %d, skipping\n", actions_read + actions_skipped + 1);
            skipping = 1;
          } else {
            /* If it worked correctly, we copy the string and terminate it. */
            strncpy(temp_command.string, &buffer[2], string_length); 
            temp_command.string[string_length] = '\0';
          }
        } else {
          robot_console_printf("Error: The string is missing in action %d, skipping\n", actions_read + actions_skipped + 1);
          skipping = 1;
        }
      } else {
        /* We split the line where there are a white spaces. */
        next_token = strtok(buffer, " ");

        /* The first token on the line should be the tag. (mendatory) */
        if (next_token != NULL && strncmp(buffer, "\n", 1) != 0) {
          temp_command.tag = next_token[0];
          next_token = strtok(NULL, " ");
        } else {
          robot_console_printf("Error: The tag is missing in action %d, skipping\n", actions_read + actions_skipped + 1);
          skipping = 1;
        }

        /* 
         * Then comes the id of the module. (mendatory)
         * There is an exception for the N command where the second argument
         * is the value which is optional
         */
        if (next_token != NULL && strncmp(buffer, "\n", 1) != 0) {
          if (temp_command.tag != 'N') {
            temp_command.roombot_id = atoi(next_token);
          } else {
            temp_command.value = atoi(next_token);
          }
          next_token = strtok(NULL, " ");
        } else if (skipping == 0) {
          if (temp_command.tag != 'N') {
            robot_console_printf("Error: The ID of the roombot is missing in action %d, skipping\n", actions_read + actions_skipped + 1);
            skipping = 1;
          }
        }

        /* Then its device id. (mendatory for the upper case commands except N) */
        if (next_token != NULL && strncmp(buffer, "\n", 1) != 0) {
          temp_command.device_id = atoi(next_token);
          next_token = strtok(NULL, " ");
        } else if (skipping == 0 && temp_command.tag < 97 && temp_command.tag != 'N') {
          robot_console_printf("Error: The device ID is missing in action %d, skipping\n", actions_read + actions_skipped + 1);
          skipping = 1;
        }

        /* And finally the value for the rotation or the behavior. (optional) */
        if (next_token != NULL && strncmp(buffer, "\n", 1) != 0) {

          /* Depending on the tag we deduce what is this token. */
          if (temp_command.tag == 'R') {
            temp_command.value = atoi(next_token);
          } else {
            temp_command.behavior = atoi(next_token);
          }
          next_token = strtok(NULL, " ");
        } else if (skipping == 0) {
            /* 
             * There is an error only for the R command as for the 
             * attaching/detaching commands only one type of module uses the 
             * behavior argument.
              */
            if (temp_command.tag == 'R') {
              robot_console_printf("Error: The value is missing in action %d, skipping\n", actions_read + actions_skipped + 1);
              skipping = 1;
            }
        }
      }
      
      /* If an error has occured, this command is not valid and we need to skip it. */
      if (skipping == 1) {
        actions_skipped++;
      } else {
        actions_read++;

        /* 
         * If the the maximal size of the array has been reached we need to 
         * resize it. 
         */
        if (actions_read > sequence_size) {
          sequence_size += SEQUENCE_SIZE;
          new_actions = realloc(actions, sequence_size * sizeof(Command));

          /* If this is impossible, we clean and return with an error value. */
          if (new_actions == NULL) {
            robot_console_printf("Error: Unable to reallocate memory for the sequence\n");
            sequence_size -= SEQUENCE_SIZE;
            clean_sequence();
            
            return 0;
          } else {
            actions = new_actions;
          }
        }
        
        /* We update the values of the current action. */
        actions[actions_read - 1].tag = temp_command.tag;
        actions[actions_read - 1].string = temp_command.string;
        actions[actions_read - 1].roombot_id = temp_command.roombot_id;
        actions[actions_read - 1].device_id = temp_command.device_id;
        actions[actions_read - 1].value = temp_command.value;
        actions[actions_read - 1].behavior = temp_command.behavior;
      }
    }

    /* We read the next line. */
    fgets(buffer, READ_BUFFER_SIZE - 1, fp);
  }

  fclose(fp);

  /* Now that we know the exact size of the array we can resize it correctly. */
  sequence_size = actions_read;
  new_actions = realloc(actions, sequence_size * sizeof(Command));

  /* In case of error, we clean everything and return with an error value. */
  if (new_actions == NULL) {
    robot_console_printf("Error: Unable to reallocate memory for the sequence\n");
    clean_sequence();
    
    return 0;
  } else {
    actions = new_actions;
  }

  return 1;
}

/* This function simply executes all the commands contained in action. */
static void execute_sequence(void)
{
  int i;

  /* 
   * Could not be simpler: for each element of the array we call the 
   * corresponding function with the arguments we stored previously.
  */
  for (i = 0; i < sequence_size; i++) {
    switch (actions[i].tag) {
    case 'A' :
      attach(actions[i].roombot_id, actions[i].device_id, actions[i].behavior);
      break;
    case 'D' :
      detach(actions[i].roombot_id, actions[i].device_id, actions[i].behavior);
      break;
    case 'R' :
      rotate(actions[i].roombot_id, actions[i].device_id, actions[i].value);
      break;
    case 'N' :
      place_new(actions[i].value);
      break;
    case 'O' :
      robot_console_printf("%s\n", actions[i].string);
      break;
    case 'a' :
      is_attached(actions[i].roombot_id);
      break;
    case 'd' :
      is_detached(actions[i].roombot_id);
      break;
    case 'r' :
      has_rotated(actions[i].roombot_id);
      break;
    default :
      robot_console_printf("Error: unknown command (#%d) %c\n", (int)actions[i].tag, actions[i].tag);
      break;
    }
  }

  return;
}

/* This function cleans up everything we made for the sequence. */
static void clean_sequence(void)
{
  int i;
  
  for (i = 0; i < sequence_size; i++) {
    if (actions[i].string != NULL) {
      free(actions[i].string);
    }
  }
  
  /* As actions was dynamically allocated we need to release this memory. */
  if (sequence_size != 0) {
    free(actions);
    sequence_size = 0;
  }
  /* Now we can either exit cleanly or reload a new sequence. */
  
  return;
}

/* This function tells to a module to attach. */
static void attach(int roombot_id, int face_id, int behavior)
{
  unsigned char buffer[4];

  buffer[0] = 'A';
  buffer[1] = roombot_id;
  buffer[2] = face_id;
  /* We always send the behavior as we do not know if it will be useful. */
  buffer[3] = behavior;

  /* We send the command on the channel of the module. */
  emitter_set_channel(emitter, roombot_id);
  emitter_send_packet(emitter, (const void *) buffer, 4);

  /*
   * We need to call this function because if there is another call to the
   * function emitter_set_channel in the same step, it will change the
   * channel for all the packets sent in this step.
   */
  robot_step(TIME_STEP);
  
  return;
}

/* This function tells to a module to detach. */
static void detach(int roombot_id, int face_id, int behavior)
{
  unsigned char buffer[4];

  buffer[0] = 'D';
  buffer[1] = roombot_id;
  buffer[2] = face_id;
  /* We always send the behavior as we do not know if it will be useful. */
  buffer[3] = behavior;

  /* We send the command on the channel of the module. */
  emitter_set_channel(emitter, roombot_id);
  emitter_send_packet(emitter, (const void *) buffer, 4);

  /*
   * We need to call this function because if there is another call to the
   * function emitter_set_channel in the same step, it will change the
   * channel for all the packets sent in this step.
   */
  robot_step(TIME_STEP);

  return;
}

/* This function tells to a module to rotate by angle (in degrees). */
static void rotate(int roombot_id, int servo_id, int angle)
{
  unsigned char buffer[5];

  buffer[0] = 'R';
  buffer[1] = roombot_id;
  buffer[2] = servo_id;
  /*
   * By using two bytes to encode the angle we can have angles 
   * from -32768 to 32767. 
   */
  buffer[3] = (angle >> 8);
  buffer[4] = (angle & 0xFF);

  /* We send the command on the channel of the module. */
  emitter_set_channel(emitter, roombot_id);
  emitter_send_packet(emitter, (const void *) buffer, 5);
  
  /*
   * We need to call this function because if there is another call to the
   * function emitter_set_channel in the same step, it will change the
   * channel for all the packets sent in this step.
   */
  robot_step(TIME_STEP);

  return;
}

/*
 * This function returns only when the corresponding module has acknowledged
 * for being attached.
 */
static void is_attached(int roombot_id)
{
  int has_done;

  has_done = 0;
  do {

    /* If there is something we read it. */
    if (receiver_get_queue_length(receiver) > 0) {
      const unsigned char *data = receiver_get_data(receiver);

      /* If this is the ack of the right module for attaching, we are done. */
      if (data[0] == 'A' && data[1] == roombot_id) {
        has_done = 1;
      } else {

      /* Else we wait to see if another packet will arrive. */
        robot_step(TIME_STEP);
      }

      /* And we erase the current packet. */
      receiver_next_packet(receiver);
    } else {

    /* Else we wait to see if a packet will arrive. */
      robot_step(TIME_STEP);
    }
  } while(has_done != 1);

  return;
}

/*
 * This function returns only when the corresponding module has acknowledged
 * for being detached.
 */
static void is_detached(int roombot_id)
{
  int has_done;

  has_done = 0;
  do {

    /* If there is something we read it. */
    if (receiver_get_queue_length(receiver) > 0) {
      const unsigned char *data = receiver_get_data(receiver);

      /* If this is the ack of the right module for detaching, we are done. */
      if (data[0] == 'D' && data[1] == roombot_id) {
        has_done = 1;
      } else {

      /* Else we wait to see if another packet will arrive. */
        robot_step(TIME_STEP);
      }
      
      /* And we erase the current packet. */
      receiver_next_packet(receiver);
    } else {
      
    /* Else we wait to see if a packet will arrive. */
      robot_step(TIME_STEP);
    }
  } while(has_done != 1);

  return;
}

/*
 * This function returns only when the corresponding module has acknowledged
 * for being in place after a rotation.
 */
static void has_rotated(int roombot_id)
{
  int has_done;

  has_done = 0;
  do {
    
    /* If there is something we read it. */
    if (receiver_get_queue_length(receiver) > 0) {
      const unsigned char *data = receiver_get_data(receiver);

      /* If this is the ack of the right module for rotating, we are done. */
      if (data[0] == 'R' && data[1] == roombot_id) {
        has_done = 1;
      } else {
        
      /* Else we wait to see if another packet will arrive. */
        robot_step(TIME_STEP);
      }
      
      /* And we erase the current packet. */
      receiver_next_packet(receiver);
    } else {
      
    /* Else we wait to see if a packet will arrive. */
      robot_step(TIME_STEP);
    }
  } while(has_done != 1);

  return;
}

/*
 * This function places a new module on the initial position. It also rotates
 * it according to angle (in degrees) on the Y axis.
 */
static void place_new(int angle)
{
  NodeRef roombot;
  char roombot_name[] = "ROOMBOT00";
  float init_values[5];

  /* 
   * The initial position is at 1 for Z. 
   * We do not modify the X and Y values.
   */
  init_values[0] = 1.0;
  /* Conversion into radians. */
  init_values[1] = 0.0;
  init_values[2] = 1.0;
  init_values[3] = 0.0;
  init_values[4] = ((float) angle * M_PI) / 180;
  /* to remember which module will be next. */
  placed_modules++;

  roombot_name[7] += (placed_modules / 10);
  roombot_name[8] += (placed_modules % 10);
  
  /* We retreive the corresponding node. */
  roombot = supervisor_node_get_from_def(roombot_name);
  /* This step is needed for the node_was_found function to work. */
  robot_step(0);
  
  /* We set the values as we want them to be. */
  if (supervisor_node_was_found(roombot) == 1) {
    supervisor_field_set(roombot, SUPERVISOR_FIELD_TRANSLATION_Z | SUPERVISOR_FIELD_ROTATION, init_values);
  } else {
    robot_console_printf("Error: next roombot to be placed not found (%s)\n", roombot_name);
  }
    
  robot_step(TIME_STEP);

  return;
}
